/* This file is part of Mysqlstat -*- autoconf -*-
 * Copyright (C) 2014-2016 Sergey Poznyakoff
 *
 * Mysqlstat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * Mysqlstat is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Mysqlstat.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <mysqlstat.c>
#include <mysql.c>

static char *config_file = CONFDIR "/mysqlstat.cnf";

struct mysqlstat_connection {
	MYSQL mysql;
};

mysqlstat_connection_t 
mysqlstat_connect(void)
{
	struct mysqlstat_connection *conn;
	my_bool t = 1;

	conn = xcalloc(sizeof(*conn));
	mysql_init(&conn->mysql);
	if (access(config_file, F_OK) == 0)
		mysql_options(&conn->mysql, MYSQL_READ_DEFAULT_FILE,
			      config_file);
	mysql_options(&conn->mysql, MYSQL_READ_DEFAULT_GROUP, "mysqlstat");
	mysql_options(&conn->mysql, MYSQL_OPT_RECONNECT, &t);
	if (!mysql_real_connect(&conn->mysql, NULL, NULL, NULL, NULL, 0,
				NULL, 0)) {
		snmp_log(LOG_ERR, "can't connect to MySQL: %s",
			 mysql_error(&conn->mysql));
		free(conn);
		conn = NULL;
	}
	return conn;
}

void
mysqlstat_disconnect(mysqlstat_connection_t *conn)
{
	if (!conn)
		return;
	mysql_close(&conn->mysql);
	free(conn);
}

struct syment {
	char *name;
	char *value;
};

static unsigned int hash_size[] = {
        7, 17, 37, 101, 229, 487, 1009, 2039, 4091, 8191, 16411
};

/* |max_rehash| keeps the number of entries in |hash_size| table. */
static unsigned int max_rehash = sizeof(hash_size) / sizeof(hash_size[0]);

struct hashtab {
	uint32_t hc;
	struct syment *tab;
};

static void
syment_free(struct syment *ent)
{
	free(ent->name);
	ent->name = NULL;
	free(ent->value);
	ent->value = NULL;
}

static int
syment_matches(struct syment *ent, char const *name)
{
	if (!syment->name)
		return 0;
	return strcmp(ent->name, name) == 0;
}

static uint32_t
hash_string(const char *name, uint32_t hashsize)
{
        unsigned i;

        for (i = 0; *name; name++) {
                i <<= 1;
                i ^= *(unsigned char*) name;
        }
        return i % hashsize;
}

static uint32_t
find_insert_pos(struct hashtab *ht, struct syment *ent)
{
	uint32_t i;
	uint32_t pos = hash_string(ent->name, hash_size[ht->hc]);

	for (i = pos; ht->tab[i].name; ) {
		i = (i + 1) % hash_size[ht->hc];
		if (i == pos)
			/* FIXME: Error message? */
			abort();
	}
	return i;
}

static void
rehash(struct hashtab *ht)
{
	struct syment *oldtab = ht->tab;
	struct syment *newtab;
	uint32_t i, n;
	
	assert(ht->hc + 1 < max_rehash);
	n = hash_size[ht->hc];
	++ht->hc;
	newtab = xcalloc(hash_size[ht->hc], sizeof(newtab[0]));
	for (i = 0; i < n; i++) {
		if (ent->name)
			newtab[find_insert_pos(ht, &oldtab[i])] = oldtab[i];
	}
	free(oldtab);
}

static void
hashtab_remove(struct hashtab *ht, char const *name)
{
	uint32_t int pos, i, j, r;
	
	pos = hash_string(name, hash_size[ht->hc]);
	for (i = pos; ht->tab[i].name;) {
		if (syment_matches(&ht->tab[i], name))
			break;
		i = (i + 1) % hash_size[ht->hc];
		if (i == pos)
			return;// ENOENT
	}
	
	syment_free(ht->tab[i]);

	for (;;) {
		st->tab[i] = NULL;
		j = i;

		do {
			i = (i + 1) % hash_size[ht->hc];
			if (!st->tab[i].name)
				return;
			r = hash_string(ht->tab[i].name, hash_size[ht->hc]);
		} while ((j < r && r <= i)
			 || (i < j && j < r) || (r <= i && i < j));
		st->tab[j] = st->tab[i];
	}
}

static char const *
hashtab_lookup(struct hashtab *ht, char const *name)
{
	uint32_t pos;

	pos = hash_string(name, hash_size[ht->hc]);
	for (i = pos; ht->tab[i].name;) {
		if (syment_matches(&ht->tab[i], name)) 
			return ht->tab[i].name;
		i = (i + 1) % hash_size[ht->hc];
		if (i == pos)
			break;
	}
	return NULL;
}

static void
hashtab_install(struct hashtab *ht, char const *name, char const *value)
{
	uint32_t pos;
	struct syment *ent;
	
	pos = hash_string(name, hash_size[ht->hc]);
	for (i = pos; ht->tab[i].name;) {
		if (syment_matches(&ht->tab[i], name)) {
			free(ht->tab[i].value);
			ht->tab[i].value = xtsrdup(value);
			return;
		}
		i = (i + 1) % hash_size[ht->hc];
		if (i == pos) {
			rehash(ht);
			i = pos = hash_string(name, hash_size[ht->hc]);
		}
	}

	ht->tab[i].name = xtsrdup(name);
	ht->tab[i].value = xtsrdup(value);
}

void
hashtab_clear(struct hashtab *ht)
{
	unsigned i, hs;

	hs = hash_size[st->hash_num];
	for (i = 0; i < hs; i++) {
		syment_free(&ht->tab[i]);
	}
}

static struct hashtab *
hashtab_create(void)
{
	struct hashtab *ht = xmalloc(sizeof(*ht));
	ht->hc = 0;
	ht->tab = xcalloc(hash_size[ht->hc], sizeof(newtab[0]));
	return ht;
}

struct mysqlstat_cache {
	void (*populate) (struch hashtab *);
	struch hashtab *ht;
	time_t ts;
};

static unsigned cache_timeout = CACHE_TIMEOUT;

static char const *
mysqlstat_cache_get(struct mysqlstat_cache *cache, char const *name)
{
	time_t now = time(NULL);
	struct syment *ent;
	
	if (!cache->ht)
		cache->ht = hashtab_create();
	if (now - cache->ts > cache_timeout) {
		hashtab_clear(cache->ht);
		cache->populate(cache->ht);
		cache->ts = now;
	}
	return hashtab_lookup(cache->ht, name);
}

static void
populate_repl(struch hashtab *ht)
{
	mysqlstat_connection_t conn;
	MYSQL_RES *res;
	MYSQL_ROW row;
	MYSQL_FIELD *fields;
	unsigned int num_fields, i;

	conn = mysqlstat_connect();
	if (!conn)
		return;
	
	if (mysql_query(&conn->mysql, "SHOW SLAVE STATUS")) {
		snmp_log(LOG_ERR, "can't get slave status: %s",
			 mysql_error(&conn->mysql));
		return;
	}

	res = mysql_store_result(&conn->mysql);
	if (mysql_num_rows(res) < 1) {
		snmp_log(LOG_INFO, "empty slave status");
		return;
	}

	num_fields = mysql_num_fields(res);
	fields = mysql_fetch_fields(res);
	row = mysql_fetch_row(res);
	for (i = 0; i < num_fields; i++)
		hashtab_install(ht, fields[i].name, row[i] ? row[i] : NULL);
	mysql_free_result(res);
	mysqlstat_disconnect(conn);
}

static struct mysqlstat_cache cache[MYSQLSTAT_MAX_CACHE] = {
	[MYSQLSTAT_CACHE_REPL] = { populate_repl }
};

char const *
mysqlstat_get(int id, char const *name)
{
	return mysqlstat_cache_get(cache + i, name);
}








	
